from __future__ import with_statement
import sys
import traceback
from time import time
import gevent

assert gevent.version_info[:2] >= (0, 13)

class Worker(gevent.Greenlet):
    """A greenlet that re-runs the function in case of an error.

    Here's an example:

        def myfunction():
            do_the_job()

        worker = Worker.spawn(myfunction)

    The above snippet is equavalent to the following:

        def myfunction():
            while True:
                try:
                    return do_the_job()
                except GreenletExit:
                    raise
                except:
                    # print the traceback to stderr, do not print the traceback if it's
                    # the same as the last time (see refresh_traceback_period)
                    # sleep for an appropriate amount (see restart_sleep and restart_rate_limit)

        gevent.spawn(myfunction)
    """
    # minimum number of seconds to sleep before restarting
    # the actual sleep time also takes restart_rate_limit in account
    restart_sleep = 5

    # (number of runs, number of seconds)
    # (5, 60) means if 5 runs happen within a minute from the start
    # of the first run then the greenlet will be put to sleep for the rest of the minute.
    restart_rate_limit = (5, 60)

    # by default if the traceback is the same as the last time, it is suppressed
    # and printed only once per this number of seconds.
    # (The one-line message that follows the traceback is printed every time).
    refresh_traceback_period = 20*60

    def __init__(self, *args, **kwargs):
        gevent.Greenlet.__init__(self, *args, **kwargs)
        self._last_tb = None
        self._last_tb_time = 0
        self.count = 0

    def run(self):
        try:
            self._start_event = None
            while True:
                sys.exc_clear()
                try:
                    self.start = time()
                    result = self._run(*self.args, **self.kwargs)
                except gevent.greenlet.GreenletExit:
                    self._report_error(sys.exc_info())
                    return
                except:
                    self._print_error(sys.exc_info())
                    self.sleep()
                    continue
                else:
                    self._report_result(result)
                    return
        finally:
            self.__dict__.pop('_run', None)
            self.__dict__.pop('args', None)
            self.__dict__.pop('kwargs', None)

    def _print_error(self, exc_info):
        tb = ''.join(traceback.format_exception(*exc_info))
        write_tb = True
        again = ''
        end = ''
        if self._last_tb == tb:
            if time() - self._last_tb_time < self.refresh_traceback_period:
                write_tb = False
            # print the whole traceback once in 10 minutes
            again = ' (again)'
        if write_tb:
            sys.stderr.write(tb)
            self._last_tb = tb
            self._last_tb_time = time()
            end = '\n'
        try:
            excname = exc_info[0].__name__
            sys.stderr.write('%s: failed with %s%s, will restart after %s\n%s' % (self, excname, again, self.get_sleep_time_as_string(), end))
        except:
            pass

    def sleep(self):
        sleep_start = time()
        while True:
            sleep_time = self.get_sleep_time() - (time() - sleep_start)
            if sleep_time <= 0:
                break
            gevent.sleep(sleep_time)
        self.count += 1
        if self.count >= self.restart_rate_limit[0]:
            self.count = 0
            self.start = time()

    def get_sleep_time(self):
        if (self.count + 1) >= self.restart_rate_limit[0]:
            period = time() - self.start
            return max(self.restart_sleep, self.restart_rate_limit[1]-period)
        else:
            return self.restart_sleep

    def get_sleep_time_as_string(self):
        return humanize_seconds(self.get_sleep_time())

    def set_sleep(self, sleep):
        self.restart_sleep = sleep

    def set_max_rate(self, max_rate):
        self.restart_rate_limit = max_rate


def humanize_seconds(period):
    """
    >>> humanize_seconds(0.1)
    '0.10 seconds'
    >>> humanize_seconds(1)
    '1 second'
    >>> humanize_seconds(89)
    '89 seconds'
    >>> humanize_seconds(60*5)
    '5:00'
    >>> humanize_seconds(60*60*3 + 2*60 + 1)
    '3:02:01'
    """
    try:
        if period < 1:
            return '%.2f second%s' % (period, 's' if period != 1 else '')
        if period < 90:
            return '%d second%s' % (round(period), 's' if period != 1 else '')
        if period < 90*60:
            minutes, seconds = (period / 60, period % 60)
            return '%d:%02d' % (minutes, seconds)
        hours, minutes, seconds = (period / 60 / 60, (period / 60) % 60, period % 60)
        return '%d:%02d:%02d' % (hours, minutes, seconds)
    except Exception:
        return '%s seconds' % (period, )


if __name__ == '__main__':
    import doctest
    doctest.testmod()

    class DieQuick(Worker):
        restart_sleep = 2
        restart_rate_limit = (5, 20)
        refresh_traceback_period = 5

        def _run(self):
            import datetime
            print datetime.datetime.now()
            1/0

        def __str__(self):
            return 'DieQuick'

    try:
        p = DieQuick.spawn()
        print 'before join'
        p.join()
        print 'after join'
    except KeyboardInterrupt:
        print 'got keyboardinterrupt!'
        p.kill()

    class SmallMaxRate(Worker):
        restart_sleep = 3
        restart_rate_limit = (2, 2)
        refresh_traceback_period = 60

        def _run(self):
            import datetime
            print datetime.datetime.now()
            1/0

        def __str__(self):
            return 'SmallMaxRate'

    try:
        p = SmallMaxRate.spawn()
        p.join()
    except KeyboardInterrupt:
        p.kill()


    class BigPeriod(Worker):
        restart_sleep = 10
        restart_rate_limit = (3, 60)
        refresh_traceback_period = 60

        def _run(self):
            import datetime
            print datetime.datetime.now()
            1/0

        def __str__(self):
            return 'BigPeriod'

    try:
        p = BigPeriod.spawn()
        p.join()
    except KeyboardInterrupt:
        print 'setting restart_sleep to 1 restart_rate_limit to (1, 1)'
        p.set_sleep(1)
        p.set_max_rate((1, 1))

    try:
        p.join()
    except KeyboardInterrupt:
        p.kill()
